from typing import Any, Tuple, Union, Iterable
import numpy as np
from pycuda.gpuarray import GPUArray

DEFAULT_TYPE: type


def predefined_read(name: str, shape: Tuple[int, ...], dtype: Any = np.float64) -> np.ndarray: ...


def tiff_read(path: str) -> np.ndarray: ...


def np_read(path: str, *, key: str = None) -> np.ndarray: ...


def mat_read(path: str, *, key: str = None) -> np.ndarray: ...


def np_write(path, arr: np.ndarray, *, scale: float = 1., pre_operator: callable = None, dtype: type = None,
             compress: bool = True): ...


def binary_write(path, arr: np.ndarray, *, scale=1., pre_operator: callable = None, dtype: type = None,
                 add_hint: bool = False): ...


def read(source: str, dtype: type = np.float64, shape: Tuple[int, ...] = None, *, scale: float = 1.,
         gpu: bool = True, **kwargs) -> Union[np.ndarray, GPUArray]: ...


def write(dest: str, array: Union[GPUArray, Iterable], **kwargs): ...



from functools import lru_cache
from pycuda import gpuarray, driver as cuda


def param_check(**kwargs):
    name0 = None
    shape0 = None
    for name in kwargs:
        arr = kwargs[name]
        if arr is None:
            continue
        # type check
        if not isinstance(arr, gpuarray.GPUArray):
            raise TypeError(f"'{name}' is not a GPUArray")
        # shape check
        if name0 is None:
            name0 = name
            shape0 = arr.shape
        else:
            if arr.shape != shape0:
                raise ValueError(f"cannot match '{name}' shape {arr.shape} with '{name0}' shape {shape0}")


@lru_cache(maxsize=None)
def get_stream(ctx):
    if ctx == cuda.Context.get_current():
        return cuda.Stream()
    else:
        raise NotImplementedError("first time can only get stream in current context. "
                                  f"ctx:{ctx}, current:{cuda.Context.get_current()}")


def get_stream_in_current():
    return get_stream(cuda.Context.get_current())


class Config:
    _res = None
    _n0 = 1.
    _xyz = None
    _lambda = None
    _callbacks = None

    def __init__(self):
        self.clear_updater()

    @property
    def xyz(self):
        if self._xyz is None:
            raise AttributeError("xyz is uninitialized")
        return self._xyz

    @property
    def lambda0(self):
        if self._lambda is None:
            raise AttributeError("wave length is uninitialized")
        return self._lambda

    @property
    def res(self):
        if self._res is None:
            raise AttributeError("res is uninitialized")
        return self._res

    @property
    def n0(self):
        return self._n0

    @res.setter
    def res(self, value):
        # if self._res is not None:
        #     warn(f"resetting res value from {self._res}")
        value = tuple(float(res_i) for res_i in value)
        if self._res != value:
            assert len(value) == 3
            self._update(attr='res', old=self._res, new=value)
            self._res = value

    @xyz.setter
    def xyz(self, value):
        self._xyz = tuple(float(size_i) for size_i in value)
        assert len(self._xyz) == 3
        self._try_calc_res()

    @lambda0.setter
    def lambda0(self, value):
        self._lambda = float(value)
        self._try_calc_res()

    @n0.setter
    def n0(self, value):
        value = float(value)
        if self._n0 != value:
            self._try_n0fix_res(value)
            self._update(attr='n0', old=self._n0, new=value)
            self._n0 = value

    def _try_calc_res(self):
        try:
            self.res = (size_i / self.lambda0 * self.n0 for size_i in self.xyz)
            # self.res = (res_i * self.n0 for res_i in self.res)
        except AttributeError:
            pass

    def _try_n0fix_res(self, new_value):
        try:
            value = (res_i / self._n0 * new_value for res_i in self.res)
            self._res = tuple(float(res_i) for res_i in value)
        except AttributeError:
            pass

    def register_updater(self, updater):
        if updater is not None:
            assert hasattr(updater, "__call__"), "updater function is not callable"
            self._callbacks.append(updater)

    def clear_updater(self):
        self._callbacks = []

    def _update(self, **kwargs):
        for i in self._callbacks:
            i(**kwargs)

    def set(self, **kwargs):
        for attr in ('n0', 'xyz', 'lambda0', 'res'):  # the order is important
            if value := kwargs.pop(attr, None):
                setattr(self, attr, value)
        # for key in kwargs:
        #     if key == "res":
        #         self.res = kwargs[key]
        #     else:
        #         raise TypeError(f"'{key}' is invalid as a configuration item")

    def __copy__(self):
        cp = type(self)()
        cp._res = self._res
        cp._n0 = self._n0
        return cp


config = Config()



import warnings


def pycuda_test():
    from pycuda.compiler import preprocess_source
    from pycuda.driver import CompileError
    from pytools.prefork import ExecError

    def warn(s):
        warnings.warn(f"pycuda doesn't seem to work normally. {s}", stacklevel=5)

    try:
        preprocess_source('', [], 'nvcc')
    except CompileError as e:
        if e.stderr:
            warn(f"nvcc basic test output: {e.stderr}")
        else:
            warn(f"maybe no available C++ compiler found by nvcc?")
    except ExecError:
        warn(f"maybe nvcc is not in PATH?")
    except:
        warn(f"unknown error")


def autoinit():
    import pycuda.driver

    if pycuda.driver.Context.get_current() is None:
        # warnings.warn("no current context")
        import pycuda.autoinit


autoinit()
pycuda_test()



import math
import functools
from warnings import warn
from collections.abc import Iterable
import numpy as np
from pycuda import gpuarray


def _cache_array(func):
    @functools.wraps(func)
    def get_cache(self, *args, gpu=False, **kwargs):
        self: Multipliers
        key, calc = func(self, *args, **kwargs)
        if gpu:
            try:
                return self._gpu_cache[key]
            except KeyError:
                arr = gpuarray.to_gpu_async(get_cache(self, *args, gpu=False, **kwargs), stream=self.stream)
                self._gpu_cache[key] = arr
                return arr
        try:
            return self._cache[key]
        except KeyError:
            arr = calc()
            assert arr.shape == self._shape
            self._cache[key] = arr
            return arr

    return get_cache


@functools.lru_cache
class Multipliers:
    def __init__(self, shape, res, stream=None):
        assert len(shape) == 2
        self._xy_size = (shape[1], shape[0])
        self._shape = shape
        self.res = res
        self._cache = {}
        self._gpu_cache = {}
        self.stream = stream

    @_cache_array
    def tilt(self, c_ab, *, trunc, periodic_params=None, c_ab_out=None):
        res = self.res
        xy_size = self._xy_size
        norm = tuple(xy_size[i] * res[i] for i in (0, 1))  # to be confirmed: * config.n0
        kernel = None
        if trunc:
            c_ab = [math.trunc(c_ab[i] * norm[i]) / norm[i] for i in (0, 1)]
        elif periodic_params is not None:
            kernel = self.gaussian(periodic_params[0], (0.5, 0.5), gpu=False)
            kernel = np.fft.fft2(np.fft.fftshift(kernel))
        c_ab = tuple(float(i) for i in c_ab)
        if c_ab_out is not None:
            for i in range(2):
                c_ab_out[i] = c_ab[i]

        key = ("t", c_ab, res, None if periodic_params is None else periodic_params[0])

        def calc():
            xr, yr = [np.arange(xy_size[i]) / xy_size[i] * c_ab[i] * norm[i] for i in (0, 1)]
            phase = np.mod(xr + yr[:, None], 1).astype(np.complex128)
            phase = np.exp(2j * np.pi * phase)
            if kernel is not None:
                phase = np.fft.fft2(phase)
                phase *= kernel
                phase = np.fft.ifft2(phase)
                # change fft default f-contiguous output to c-contiguous
                phase = np.ascontiguousarray(phase)
            # normalize by center point value
            phase /= phase[tuple(i // 2 for i in phase.shape)]
            np.testing.assert_almost_equal(phase[tuple(i // 2 for i in phase.shape)], 1)
            return phase

        return key, calc

    @_cache_array
    def binary_pupil(self, na):
        res = self.res
        c_gamma = self.c_gamma()
        key = ("bp", round(na * 1000), res)

        def calc():
            mask = np.greater(c_gamma, np.sqrt(1 - na ** 2))
            mask = mask.astype(np.complex128)
            return mask

        return key, calc

    @_cache_array
    def c_gamma(self):
        """
        Calculate cos(gamma) as a constant array at frequency domain. Gamma is the angle
        between the wave vector and z-axis. Note: N.A.=sin(gamma)

        The array is pre-shifted for later FFT operation.

        :return: cos(gamma) numpy array
        """
        xy_size = self._xy_size
        res = self.res
        key = ("cg", res)

        def calc():
            eps = 1E-8
            c_alpha, c_beta = [
                np.fft.ifftshift(np.arange(-xy_size[i] / 2, xy_size[i] / 2).astype(np.double)) / xy_size[i] / res[i]
                for i in (0, 1)
            ]
            c_gamma = np.sqrt(np.maximum(1 - (np.square(c_alpha) + np.square(c_beta[:, None])), eps))
            return c_gamma

        return key, calc

    @staticmethod
    def _near_0(size, pos_0):
        """
        Return ``f(x)``, which satisfies:

        ``f(x)`` range ``[-0.5, 0.5)``,

        ``f'(x) = 1`` (except for the discontinuity point jumping from 0.5 to -0.5),

        ``f(pos_0) = 0``

        :param size: arr length
        :param pos_0: position of zero
        :return:
        """
        return np.mod((np.arange(size).astype(np.double) + 0.5) / size + 0.5 - pos_0, 1) - 0.5

    @_cache_array
    def soft_crop(self, width, *, total_slices=1, pos=0, strength=1):
        xy_size = self._xy_size
        width = float(width)
        if width >= 1 or width <= 0:
            raise ValueError("width should be a relative value in 0-1")
        key = ("cr", round(width * 100), round(pos * 100), round(total_slices), round(strength * 100))
        if strength < 0.01:
            warn("strength is too weak, change to 0.01", stacklevel=3)
            strength = 0.01

        def calc():
            x, y = [
                np.exp(-np.exp(-(self._near_0(xy_size[i], pos) * 2 / width) ** 2 * (np.log(100 * strength) + 0.8))
                       * 100 * strength / total_slices)
                for i in (0, 1)
            ]
            mask = x * y[:, None]
            return mask

        return key, calc

    @_cache_array
    def hard_crop(self, width):
        """
        Possible width specs:

        * 0~1 ``float``: relative crop value, same for each axis
        * ``int``: crop pixels, same for each axis
        * ``Iterable`` (list, tuple, ...): Indicate crop pixels of ``(left, right, top, bottom)``,
          or ``(x, y)`` for same pixels each side

        :param width: crop size specification
        """
        err_width = ValueError(f"invalid width spec {width}")
        xy_size = self._xy_size
        if isinstance(width, float):
            width = width / 2 + 1e-8
            x_size, y_size = [int(width * i) for i in xy_size]
            width = (x_size, x_size, y_size, y_size)
        elif isinstance(width, int):
            width = (width,) * 4
        elif isinstance(width, Iterable):
            width = tuple(width)
            if len(width) == 2:
                x_size, y_size = width
                width = (x_size, x_size, y_size, y_size)
            elif len(width) != 4:
                raise err_width
        else:
            raise err_width
        for i in width:
            if not (isinstance(i, int) and i >= 0):
                raise err_width
        key = ("hc", width)

        def calc():
            x, y = [
                np.zeros(xy_size[i], bool)
                for i in (0, 1)
            ]
            x1, x2, y1, y2 = width
            x[x1:-x2] = True
            y[y1:-y2] = True
            mask = np.logical_and(x, y[:, None])
            return mask.astype(np.double)

        return key, calc

    @_cache_array
    def gaussian(self, sigma, mu=(0, 0)):
        xy_size = self._xy_size
        sigma = float(sigma)
        mu = tuple(float(i) for i in mu)
        for i in mu:
            if i > 1 or i < 0:
                warn("mu value is out of normal range 0~1")
        key = ("ga", round(sigma * 100), tuple(round(i * 100) for i in mu))

        def calc():
            x, y = [
                np.exp(-(self._near_0(xy_size[i], mu[i])) ** 2
                       / 2 / (sigma / xy_size[i]) ** 2)
                for i in (0, 1)
            ]
            mask = x * y[:, None]
            return mask

        return key, calc



import sys

import numpy as np
from tifffile import TiffWriter, TiffFile
from scipy.io import loadmat, whosmat, savemat
from warnings import warn
import os
import csv
from pycuda import gpuarray
from pycuda import driver as cuda

DEFAULT_TYPE = np.float64


# if raw_img.shape != SIZE[:2]:
#     raise ValueError(f"Input image size {raw_img.shape} is incompatible"
#                      f"with x-y size {SIZE[:2]}")

def predefined_read(name, shape, dtype=None):
    """
    Get a predefined tensor
    ``shape`` can be a ``list`` of integers, a ``tuple`` of integers,
    or a 1-D ``Tensor`` of type ``int32``

    :param name: predefined image name
    :param shape: Dimensions of resulting tensor
    :param dtype: DType of the elements of the resulting tensor
    """
    if shape is None:
        raise ValueError("indeterminate shape")
    if name == "plane":
        img = np.ones(shape, dtype)
    else:
        raise ValueError(f"unknown image name {name}")
    return img


def tiff_read(path):
    """
    Import a TIFF file as numpy array

    :param path: Target file path
    :return: A tensor constant
    """
    with TiffFile(path) as file:
        img = np.squeeze(np.stack([page.asarray() for page in file.pages]))
    if img.dtype.type == np.uint16:
        img = img.astype(np.double) / 65535
    elif img.dtype.type == np.uint8:
        warn("Importing uint8 image, please use uint16 if possible")
        img = img.astype(np.double) / 255
    else:
        raise TypeError(f"Unknown data type {img.dtype.type} of input image")
    return img


def np_read(path, *, key=None):
    ext = os.path.splitext(path)[-1]
    if key is None:
        key = 'arr_0'
    if ext == '.npy':
        img = np.load(path)
    elif ext == '.npz':
        with np.load(path) as f:
            try:
                img = f[key]
            except KeyError:
                warn(f"'{key}' is not a file in archive {path}. Reading first file instead.", stacklevel=2)
                img = f[f.files[0]]
    else:
        raise ValueError(f"unknown filename extension '{ext}'")
    return img


def csv_read(path: str):
    table = []
    with open(path, 'r', newline='') as file:
        reader = csv.reader(file, quoting=csv.QUOTE_NONNUMERIC)
        for row in reader:
            table.append(row)
    return np.array(table)


def mat_read(path, *, key=None):
    if key is None:
        key = whosmat(path)[0][0]
    img = loadmat(path)[key]
    if not isinstance(img, np.ndarray):
        raise TypeError(f"Unknown data type {type(img)} of input image")
    return img


def read(source, dtype=None, shape=None, *, scale=1., gpu=True, pagelocked=False, **kwargs):
    """
    Read a ``tf.Tensor`` from source. Method is auto-detected corresponding to the file extension.

    :param source: Target file path
    :param dtype: Data type of the elements of the resulting tensor
    :param shape:
    :param scale:
    :param gpu: whether load memory to gpu
    :param pagelocked: for cpu, whether make it pinned
    :return:
    """
    if source in {'plane'}:
        arr = predefined_read(source, shape, dtype)
    elif os.access(source, os.R_OK):
        ext = os.path.splitext(source)[-1]
        if ext in {'.tiff', '.tif'}:
            arr = tiff_read(source)
        elif ext in {'.npy', '.npz'}:
            arr = np_read(source, **kwargs)
        elif ext == '.mat':
            arr = mat_read(source, **kwargs)
        elif ext == '.csv':
            arr = csv_read(source)
        else:
            raise ValueError(f"unknown filename extension '{ext}'")
        if dtype is not None:
            arr = arr.astype(dtype, copy=False)

    else:
        raise FileNotFoundError(f"'{source}' is not a known illumination type nor a valid file path")
    if shape is not None and tuple(shape) != tuple(arr.shape):
        arr = arr.reshape(shape)
    rank = len(arr.shape)
    if rank not in {2, 3}:
        warn(f"Importing {rank}-D image with shape {arr.shape}", stacklevel=2)
    arr *= scale
    if gpu:
        arr = gpuarray.to_gpu(arr)
    else:
        if pagelocked:
            arr = cuda.register_host_memory(arr)
    return arr


def tiff_write(path, arr, *, scale=1, pre_operator: callable = None, dtype=np.uint16,
               compress: bool = True, photometric=None):
    """
    Export a list of Tensors to a multi-page tiff

    ``pre_operator`` can apply some numpy functions before the data to be exported, such as brightness
    and contrast adjustment. Normally it is used to avoid saturation or too dark pictures

    Argument ``scale`` is a simpler way than `pre_operator` to adjust data range.
    If ``scale`` is ``None``, it will write the raw data. Please note that this is
    **different** with ``scale=1``

    Argument ``dtype`` means the color depth of the exported image. It must be 16bit - ``np.uint16``
    (default) or 8bit - ``np.uint8``

    :param path: Target file path
    :param arr: Tensor data to be written
    :param scale: Multiplier to adjust value range
    :param pre_operator: Preprocess function before scaling
    :param dtype: Color depth of the exported image
    :param compress: Using lossless compression
    :param photometric:
    """
    arr = np.squeeze(arr)
    shape = arr.shape
    if photometric == 'rgb':
        assert shape[-1] == 3, "rgb need exactly 3 channels"
        shape = shape[:-1]
    if len(shape) == 2:
        arr = [arr]
    elif len(shape) != 3:
        raise ValueError(f"Must export 2-D or 3-D array but got {len(arr.shape)}-D data, "
                         f"shape {arr.shape}.")
    with TiffWriter(path) as out_file:
        for i in arr:
            if pre_operator is not None:
                i = pre_operator(i)
            try:
                if scale is not None:
                    i = i * scale * {np.uint16: 65535, np.uint8: 255}[dtype]
                i = i.astype(np.int64)
                np.clip(i, 0, {np.uint16: 65535, np.uint8: 255}[dtype], out=i)
            except KeyError:
                raise TypeError(f"dtype should be either np.uint8 or np.uint16, but not {dtype}") from None
            i = i.astype(dtype)
            if compress:
                out_file.save(i, compress=9, predictor=True, photometric=photometric)
            else:
                out_file.save(i, photometric=photometric)


def np_write(path, arr, *, scale=1., pre_operator=None, dtype=None, compress=True):
    if pre_operator is not None:
        arr = pre_operator(arr)
    arr = arr * scale
    if dtype is not None:
        arr = arr.astype(dtype)
    ext = os.path.splitext(path)[-1]
    if ext == '.npy':
        np.save(path, arr)
    elif ext == '.npz':
        np.savez_compressed(path, arr) if compress else np.savez(path, arr)


def binary_write(path, arr, *, scale=1., pre_operator=None, dtype=None, add_hint=False):
    if pre_operator is not None:
        arr = pre_operator(arr)
    arr = arr * scale
    if dtype is not None:
        arr = arr.astype(dtype)
    if add_hint:
        path = os.path.splitext(path)
        path_list = [path[0]]
        for i in arr.shape:
            path_list.append(f"_{i}")
        path_list.append(f"_{arr.dtype}")
        path_list.append(f"_{sys.byteorder[0]}e")
        path_list.append(path[1])
        path = "".join(path_list)
    arr.tofile(path)


def csv_write(path: str, table):
    with open(path, 'w', newline='') as file:
        writer = csv.writer(file)
        for row in table:
            writer.writerow(row)


def mat_write(path, arr, *, scale=1., pre_operator=None, dtype=None, compress=True, key="data"):
    if pre_operator is not None:
        arr = pre_operator(arr)
    arr = arr * scale
    if dtype is not None:
        arr = arr.astype(dtype)
    savemat(path, {key: arr}, do_compression=compress)


def write(dest, array, **kwargs):
    if isinstance(array, np.ndarray):
        arr = array
    elif isinstance(array, gpuarray.GPUArray):
        arr = array.get()
    else:
        arr = [page for page in array]
        if isinstance(arr[0], np.ndarray):
            arr = np.stack(arr)
        elif isinstance(arr[0], gpuarray.GPUArray):
            arr = np.stack([gpu_arr.get() for gpu_arr in arr])
        else:
            try:
                warn("trying general np.array fallback", stacklevel=2)
                arr = np.array(array, copy=False)
            except Exception as ee:
                raise TypeError("unknown data type to write") from ee

    ext = os.path.splitext(dest)[-1]
    if ext in {'.tiff', '.tif'}:
        tiff_write(dest, arr, **kwargs)
    elif ext in {'.npy', '.npz'}:
        np_write(dest, arr, **kwargs)
    elif ext == '.mat':
        mat_write(dest, arr, **kwargs)
    elif ext == '.bin':
        binary_write(dest, arr, **kwargs)
    elif ext == '.csv':
        csv_write(dest, arr)
    else:
        raise ValueError(f"unknown filename extension '{ext}'")



from typing import Tuple, Literal, Union
from pycuda.driver import Stream
from ssnp.funcs import BPMFuncs, SSNPFuncs, Funcs
from pycuda.gpuarray import GPUArray
import numpy as np
from numbers import Real
# from .utils import Multipliers
from ssnp.utils import Config


def ssnp_step(u: GPUArray, u_d: GPUArray, dz: Real, n: GPUArray = None, output: GPUArray = None,
              config: Config = None, stream: Stream = None) -> Tuple[GPUArray, GPUArray]: ...


def bpm_step(u: GPUArray, dz: Real, n: GPUArray = None, output: GPUArray = None,
             config: Config = None, stream: Stream = None) -> GPUArray: ...


def bpm_grad_bp(u: GPUArray, ug: GPUArray, dz: Real, n: GPUArray = None, ng=None,
                config: Config = None, stream: Stream = None) -> GPUArray: ...


def ssnp_grad_bp(u, ug, u_dg, dz, n=None, ng=None,
                 config: Config = None, stream: Stream = None) -> GPUArray: ...


def reduce_mse(u: GPUArray, measurement: GPUArray, stream: Stream = None) -> np.double: ...


def reduce_mse_grad(u: GPUArray, measurement: GPUArray, output: GPUArray = None, stream: Stream = None) -> GPUArray: ...


# def pure_forward_d(u: GPUArray, output: GPUArray = None) -> GPUArray: ...


# def binary_pupil(u: GPUArray, na: float, multiplier: Multipliers = None) -> GPUArray: ...

def sum_batch(u: GPUArray, output: GPUArray = None, stream: Stream = None): ...


def get_multiplier(shape, res=None, stream: Stream = None): ...


def u_mul(u: GPUArray, mul, copy: bool = False, stream: Stream = None, conj: bool = False) -> GPUArray: ...


def merge_prop(uf: GPUArray, ub: GPUArray, config: Config = None,
               copy: bool = False, stream: Stream = None) -> Tuple[GPUArray, GPUArray]: ...


def split_prop(u: GPUArray, u_d: GPUArray, config: Config = None,
               copy: bool = False, stream: Stream = None) -> Tuple[GPUArray, GPUArray]: ...


def merge_grad(ufg: GPUArray, ubg: GPUArray, config: Config = None,
               copy: bool = False, stream: Stream = None) -> Tuple[GPUArray, GPUArray]: ...


def get_funcs(arr_like: GPUArray, config: Config = None, *,
              model: Literal['ssnp', 'bpm', 'any', 'BPM', 'SSNP', 'Any', 'ANY'] = 'any',
              stream: Stream = None,
              fft_type: Literal["reikna", "skcuda"] = "skcuda") -> Union[BPMFuncs, SSNPFuncs, Funcs]: ...



"""
Module for BeamArray class
"""
from warnings import warn
import copy
from collections import Iterable
from numbers import Number
import numpy as np
from pycuda import gpuarray
from pycuda.gpuarray import GPUArray
from ssnp import calc
from ssnp.utils import param_check, Config
import logging


class BeamArray:
    """
    BeamArray provide convenient ways to perform operations for GPUArray (or GPUArrays pair):

    1. For SSNP, forward/backward and field/derivation representation can convert transparently when needed.

    2. Operation can be tracked for ``n_grad`` to get n gradient easily and fast
    """
    dtype = np.complex128
    DERIVATIVE = 0
    BACKWARD = 1
    _config = None

    def __init__(self, u1, u2=None, relation=BACKWARD, total_ops=0, stream=None):
        def to_gpu(u, name):
            if isinstance(u, (np.ndarray, GPUArray)):
                if u.dtype != self.dtype:
                    warn(f"force casting {name} to {self.dtype}", stacklevel=3)
                    u = u.astype(self.dtype)
            else:
                u = np.array(u, dtype=self.dtype)
                warn(f"converting {name} to numpy array as fallback", stacklevel=3)
            return gpuarray.to_gpu_async(u, stream=stream)  # async version of copy()

        self._u1 = to_gpu(u1, "u1")
        shape = self._u1.shape
        if len(shape) == 2:
            self.batch = 1
        elif len(shape) == 3:
            self.batch = shape[0]
        else:
            raise ValueError(f"cannot process {len(shape)}-D data with shape {shape}")

        if u2 is not None:
            self._u2 = to_gpu(u2, "u2")
            if self._u2.shape != shape:
                raise ValueError(f"u1 shape {shape} cannot match u2 shape {self._u2.shape}")
            if relation not in {BeamArray.DERIVATIVE, BeamArray.BACKWARD}:
                raise ValueError("unknown relation type")
            self.relation = relation
        else:
            self._u2 = None
        self._tape = []
        self._array_pool = []
        self.shape = shape[-2:]
        # self.multiplier = calc.get_multiplier(shape, stream=stream)
        self._get_array_times = 0

        max_ops = int(np.sqrt(2 * total_ops)) + 1 if total_ops > 0 else 0
        self.ops_number = {"max": max_ops, "remainder": max_ops, "current": max_ops}
        self._fft_funcs = calc.get_funcs(self._u1, stream=stream)
        self.stream = stream

    @property
    def multiplier(self):
        res = None if self._config is None else self._config.res
        return calc.get_multiplier(self.shape, res, stream=self.stream)

    @property
    def config(self):
        if self._config is None:
            self._config = Config()
            self.register_config_updater()

        return self._config

    @config.setter
    def config(self, value):
        assert isinstance(value, Config)
        self._config = copy.deepcopy(value)
        self._config.clear_updater()
        self.register_config_updater()

    def register_config_updater(self):
        def update(attr, **_):
            if self._u2 is not None:
                if attr == 'res':
                    self.split_prop()
                elif attr == 'n0':
                    self.merge_prop()

        self._config.register_updater(update)

    def _get_array(self):
        if self._array_pool:
            return self._array_pool.pop()
        else:
            self._get_array_times += 1
            logging.info(f"get array times: {self._get_array_times}")
            return gpuarray.empty_like(self._u1)

    def recycle_array(self, arr):
        param_check(beam=self._u1, recycle=arr)
        assert arr.dtype == self.dtype
        self._array_pool.append(arr)

    def split_prop(self):
        if self._u2 is None:
            warn("split_prop for forward-only beam is useless")
            return
        if self.relation == BeamArray.DERIVATIVE:
            calc.split_prop(self._u1, self._u2, self._config, stream=self.stream)
            self.relation = BeamArray.BACKWARD

    def merge_prop(self):
        if self._u2 is None:
            warn("merge_prop for forward-only beam is useless")
            return
        if self.relation == BeamArray.BACKWARD:
            calc.merge_prop(self._u1, self._u2, self._config, stream=self.stream)
            self.relation = BeamArray.DERIVATIVE

    @property
    def forward(self):
        if self._u2 is not None:
            self.split_prop()
        return self._u1

    @property
    def backward(self):
        if self._u2 is not None:
            self.split_prop()
        return self._u2

    @property
    def field(self):
        if self._u2 is not None:
            self.merge_prop()
        return self._u1

    @property
    def derivative(self):
        if self._u2 is not None:
            self.merge_prop()
        return self._u2

    @staticmethod
    def _u_setter(u, value, dtype=dtype):
        if isinstance(value, (GPUArray, np.ndarray)):
            u.set(value)
        elif isinstance(value, Number):
            value: float
            u.fill(np.array(value, dtype=dtype))
        else:
            raise AttributeError(f"cannot set value type {type(value)} to BeamArray")

    @forward.setter
    def forward(self, value):
        if self._u2 is not None:
            self.split_prop()
        self._u_setter(self._u1, value)

    @backward.setter
    def backward(self, value):
        if self._u2 is not None:
            self.split_prop()
        if value is None:
            if self._u2 is not None:
                self.recycle_array(self._u2)
                self._u2 = None
        else:
            if self._u2 is None:
                self._u2 = self._get_array()
                self.relation = BeamArray.BACKWARD
            self._u_setter(self._u2, value)

    @field.setter
    def field(self, value):
        if self._u2 is not None:
            self.merge_prop()
        self._u_setter(self._u1, value)

    @derivative.setter
    def derivative(self, value):
        if self._u2 is not None:
            self.merge_prop()
        if value is None:
            if self._u2 is not None:
                self.recycle_array(self._u2)
                self._u2 = None
        else:
            if self._u2 is None:
                self._u2 = self._get_array()
                self.relation = BeamArray.DERIVATIVE
            self._u_setter(self._u2, value)

    def ssnp(self, dz, n=None, *, track=False):
        """
        Perform an SSNP operation for this beam array

        :param dz: z slice thickness, unit: z pixel size
        :param n: refractive index, default is background n0
        :param track: track this operation for gradient calculation
        """
        if self._u2 is None:
            warn("incomplete field information, assuming u is pure forward")
            self.backward = 0
        self.merge_prop()
        self._parse(('ssnp', self._u1, self._u2), dz, n, track)

    def bpm(self, dz, n=None, *, track=False):  # todo: untested
        """
        Perform a BPM operation for this beam array

        :param dz: z slice thickness, unit: z pixel size
        :param n: refractive index, default is background n0
        :param track: track this operation for gradient calculation
        """
        if self._u2 is not None:
            warn("discarding backward propagation part of bidirectional field", stacklevel=2)
            self.backward = None
        self._parse(('bpm', self._u1), dz, n, track)

    def n_grad(self, output=None):  # todo: correct config tracking
        # output shape checking/setting
        scatter_num = 0
        ug = None
        u_dg = None
        ng_batch = gpuarray.empty_like(self._u1, dtype=np.float64)
        for op in self._tape:
            if op[0] in {'ssnp', 'bpm'} and op[-1] is not None:
                scatter_num += 1
        if output is None:
            # new is scatter_num times of empty_like with double dtype
            output = GPUArray((scatter_num, *self._u1.shape[-2:]), np.double)
        elif len(output) != scatter_num:
            raise ValueError(f"output length {len(output)} is different from scatter operation number {scatter_num}")

        # processing
        while len(self._tape) > 0:
            op = self._tape.pop()
            if op[0] == "mse":
                if ug is not None:
                    raise ValueError("multiple reduce operation")
                ug = self._get_array()
                calc.reduce_mse_grad(self.forward, op[1], output=ug)

            elif op[0] == "bpm":
                if ug is None:
                    raise ValueError("no reduce operation")
                _, u, dz, n = op
                if n is None:
                    calc.bpm_grad_bp(u, ug, dz)
                else:
                    if u is None:
                        index = len(self._tape) - 1
                        while True:
                            if self._tape[index][1] is None:
                                index = index - 1
                                assert index >= 0
                            else:
                                if self._tape[index][0] == "bpm":
                                    break
                        index_u = self._get_array()
                        index_u.set(self._tape[index][1])
                        index += 1
                        # recalculate previous u for n!=None but u==None ops
                        while index < len(self._tape):
                            if self._tape[index][0] == "bpm":
                                calc.bpm_step(index_u, *self._tape[index][2:])
                                if self._tape[index][3] is not None:
                                    assert self._tape[index][1] is None
                                    self._tape[index][1] = index_u
                                    index_u = self._get_array()
                                    index_u.set(self._tape[index][1])
                            index += 1
                        calc.bpm_step(index_u, dz, n)  # recalculate current step u
                        u = index_u
                    # u, ug, dz, n is all checked and not None
                    scatter_num -= 1
                    if self.batch > 1:
                        calc.bpm_grad_bp(u, ug, dz, n, ng_batch)
                        calc.sum_batch(ng_batch, output[scatter_num])
                    else:
                        calc.bpm_grad_bp(u, ug, dz, n, output[scatter_num])
                    self.recycle_array(u)

            elif op[0] == "ssnp":
                if ug is None:
                    raise ValueError("no reduce operation")
                _, u, _, dz, n = op
                if u_dg is None and ug is not None:
                    ufg = ug
                    ubg = self._get_array()
                    ubg.fill(np.array(0, self.dtype))
                    u, u_dg = calc.merge_grad(ufg, ubg)
                if n is None:
                    calc.ssnp_grad_bp(u, ug, u_dg, dz)
                else:
                    scatter_num -= 1
                    if self.batch > 1:
                        calc.ssnp_grad_bp(u, ug, u_dg, dz, n, ng_batch)
                        calc.sum_batch(ng_batch, output[scatter_num])
                    else:
                        calc.ssnp_grad_bp(u, ug, u_dg, dz, n, output[scatter_num])
                    self.recycle_array(u)

            elif op[0] == "u*":
                _, mul = op
                calc.u_mul(ug, mul, conj=True)
                if u_dg is not None:
                    calc.u_mul(u_dg, mul, conj=True)

            elif op[0] == "a*":
                _, mul = op
                with self._fft_funcs.fourier(ug):
                    calc.u_mul(ug, mul, conj=True)
                if u_dg is not None:
                    with self._fft_funcs.fourier(u_dg):
                        calc.u_mul(u_dg, mul, conj=True)
            else:
                raise NotImplementedError(f"unknown operation {op[0]}")
        assert scatter_num == 0
        if ug is not None:
            self.recycle_array(ug)
        if u_dg is not None:
            self.recycle_array(u_dg)
        self.ops_number["remainder"] = self.ops_number["current"] = self.ops_number["max"]
        return output

    def binary_pupil(self, na):
        self.a_mul(self.multiplier.binary_pupil(na, gpu=True))

    def mul(self, arr, *, hold=None, track=False):
        """
        calculate ``beam *= arr``

        If hold is given, ``beam = (beam - hold) * arr + hold``

        :param arr: other array to multiply
        :param hold: some part not perform multiply
        :param track: track this operation for gradient calculation
        """
        param_check(field=self._u1, multiplier=arr, hold=hold and hold._u1)
        if hold is not None:
            self.__isub__(hold)
            self.mul(arr, track=track)
            self.__iadd__(hold)
        else:
            self.__imul__(arr)
            if track:
                self._tape.append(("u*", arr))

    def a_mul(self, arr, hold=None, track=False):
        fourier = self._fft_funcs.fourier
        if self.batch == 1:
            param_check(angular_spectrum=self._u1, multiplier=arr, hold=None if hold is None else hold._u1)
        else:
            param_check(angular_spectrum=self._u1[0], multiplier=arr)
        if hold is not None:
            self.__isub__(hold)
            self.a_mul(arr, track=track)
            self.__iadd__(hold)
        else:
            with fourier(self._u1):
                calc.u_mul(self._u1, arr)
            if self._u2 is not None:
                with fourier(self._u2):
                    calc.u_mul(self._u2, arr)
            if track:
                self._tape.append(("a*", arr))

    def __imul__(self, other):  # todo: move to calc
        calc.u_mul(self._u1, other, stream=self.stream)
        if self._u2 is not None:
            calc.u_mul(self._u2, other, stream=self.stream)
        return self

    @staticmethod
    def _iadd_isub(self, other, add):  # todo async
        assert isinstance(other, BeamArray)
        param_check(self=self._u1, add=other._u1)
        if self._u2 is None:
            if other._u2 is None:
                if add:
                    self._u1 += other._u1
                else:
                    self._u1 -= other._u1
            else:
                raise ValueError("incompatible BeamArray type")
        else:
            if other._u2 is None:
                raise ValueError("incompatible BeamArray type")
            else:
                if self.relation == BeamArray.DERIVATIVE:
                    other.merge_prop()
                else:
                    other.split_prop()
                if add:
                    self._u1 += other._u1
                    self._u2 += other._u2
                else:
                    self._u1 -= other._u1
                    self._u2 -= other._u2
        return self

    def __iadd__(self, other):
        return BeamArray._iadd_isub(self, other, add=True)

    def __isub__(self, other):
        return BeamArray._iadd_isub(self, other, add=False)

    def forward_mse_loss(self, measurement, *, track=False):
        if track:
            self._tape.append(("mse", measurement))
        return calc.reduce_mse(self.forward, measurement)

    def _parse(self, info, dz, n, track):
        def step(var_dz, var_n):
            if info[0] == 'bpm':
                calc.bpm_step(info[1], var_dz, var_n, config=self._config)
                if track:
                    if var_n is None:
                        u_save = None
                    else:
                        if self.ops_number["current"] >= self.ops_number["remainder"]:
                            u_save = self._get_array()
                            u_save.set(info[1])
                            if self.ops_number["remainder"] > 0:
                                self.ops_number["remainder"] -= 1
                                self.ops_number["current"] = 0
                        else:
                            self.ops_number["current"] += 1
                            u_save = None
                    self._tape.append(['bpm', u_save, var_dz, var_n])
            elif info[0] == 'ssnp':
                calc.ssnp_step(info[1], info[2], var_dz, var_n, config=self._config)
                if track:
                    if var_n is None:
                        u_save = None
                    else:
                        u_save = self._get_array()
                        u_save.set(info[1])
                    # ud_save = self._get_array()
                    # ud_save.set(info[2])
                    ud_save = None
                    self._tape.append(('ssnp', u_save, ud_save, var_dz, var_n))

        if n is None:
            if isinstance(dz, Iterable):
                for dzi in dz:
                    step(dzi, None)
            else:
                step(dz, None)
        else:
            if isinstance(n, GPUArray):
                if len(n.shape) == 2:
                    step(dz, n)
                    return
                elif len(n.shape) != 3:
                    raise ValueError(f"invalid n shape {n.shape}")
            if isinstance(dz, Iterable):
                for dz_n in zip(dz, n):
                    step(*dz_n)
            else:
                for ni in n:
                    step(dz, ni)

    def __del__(self):
        # for arr in self._array_pool:
        #     del arr
        del self._u1
        del self._u2

    def __repr__(self):
        if self._u2:
            if self.relation == BeamArray.DERIVATIVE:
                beam_type = "derivative"
                data = ("total complex field U", "diff[U(z), z]")
            elif self.relation == BeamArray.BACKWARD:
                beam_type = "backward"
                data = ("forward component Uf", "backward component Ub")
            else:
                raise TypeError("unknown beam type")

            beam_type = "bi-directional beam, additional data is " + beam_type
            data = f"{{{data[0]}:\n{self._u1},\n{data[1]}:\n{self._u2}}}"
        else:
            beam_type = "forward-only beam"
            data = f"{{complex field U:\n{self._u1}}}"

        return f"{{size: {self.shape},\nbeam type: {beam_type},\ndata: {data}\n}}"


class Operation:
    def __init__(self, forward, gradient, name):
        self.type


class OperationTape:
    _ops = None

    @staticmethod
    def tally(total):
        if total <= 0:
            while True:
                yield True
        remainder = current = int(np.sqrt(2 * total + 0.25) - 0.5)
        while True:
            if current >= remainder:
                if remainder > 0:
                    remainder -= 1
                    current = 0
                yield True
            else:
                current += 1
                yield False

    def __init__(self, size):
        self._tape = []
        self.need_save = self.tally(size)

    def append(self, op):
        assert isinstance(op, Operation)





import numpy as np
from pycuda import gpuarray
from ssnp.funcs import BPMFuncs, SSNPFuncs, Funcs
from ssnp.utils import param_check, config as global_config, Multipliers
from numbers import Number


def ssnp_step(u, u_d, dz, n=None, output=None, config=None, stream=None):
    """
    SSNP main operation of one step

    :param u: x-y complex amplitude
    :param u_d: z partial derivative of u
    :param dz: step size along z axis
    :param n: refractive index along x-y distribution in this slice. Use background N0 if not provided.
    :param output:
    :param config:
    :param stream:
    :return: new (u, u_d) after a step towards +z direction
    """
    if len(u.shape) == 3:
        param_check(u_batch=u, u_d_batch=u_d, output_batch=output)
        param_check(u=u[0], n=n)
    else:
        param_check(u=u, u_d=u_d, n=n, output=output)
    funcs: SSNPFuncs = get_funcs(u, config, model="ssnp", stream=stream)
    a = funcs.fft(u, output=output)
    a_d = funcs.fft(u_d, output=output)
    funcs.diffract(a, a_d, dz)
    u = funcs.ifft(a)
    u_d = funcs.ifft(a_d)
    # n = n / N0
    if n is not None:
        funcs.scatter(u, u_d, n, dz)
    return u, u_d


def bpm_step(u, dz, n=None, output=None, config=None, stream=None):
    """
    BPM main operation of one step

    :param u: x-y complex amplitude
    :param dz: step size along z axis
    :param n: refractive index along x-y distribution in this slice. Use background N0 if not provided.
    :param output:
    :param config:
    :param stream:
    :return: new (u, u_d) after a step towards +z direction
    """
    if len(u.shape) == 3:
        param_check(u_batch=u, output_batch=output)
        param_check(u=u[0], n=n)
    else:
        param_check(u=u, n=n, output=output)
    funcs: BPMFuncs = get_funcs(u, config, model="bpm", stream=stream)
    a = funcs.fft(u, output=output)
    funcs.diffract(a, dz)
    u = funcs.ifft(a)
    if n is not None:
        funcs.scatter(u, n, dz)
    return u


def bpm_grad_bp(u, ug, dz, n=None, ng=None, config=None, stream=None):
    param_check(u_1=u, u_grad=ug, n_grad=ng)
    param_check(u_1=u[0] if len(u.shape) == 3 else u, n=n)
    funcs: BPMFuncs = get_funcs(ug, config, model="bpm", stream=stream)
    if n is not None:
        funcs.scatter_g(u, n, ug, ng, dz)
    with funcs.fourier(ug) as ag:
        funcs.diffract_g(ag, dz)
    return ng


def ssnp_grad_bp(u, ug, u_dg, dz, n=None, ng=None, config=None, stream=None):
    param_check(u_1=u, u_grad=ug, u_d_grad=u_dg, n_grad=ng)
    param_check(u_1=u[0] if len(u.shape) == 3 else u, n=n)
    funcs: SSNPFuncs = get_funcs(ug, config, model="ssnp", stream=stream)
    if n is not None:
        funcs.scatter_g(u, n, ug, u_dg, ng, dz)
    with funcs.fourier(ug) as ag, funcs.fourier(u_dg) as a_dg:
        funcs.diffract_g(ag, a_dg, dz)
    return ng


def reduce_mse(u, measurement, stream=None):
    param_check(u=u, measurement=measurement)
    if u.dtype != np.complex128:
        raise ValueError(f"u dtype {u.dtype} is incompatible")
    funcs = get_funcs(u, model="any", stream=stream)
    # if measurement.dtype == np.complex128:
    #     result = funcs.reduce_mse_cc(u, measurement)
    # elif measurement.dtype == np.double:
    #     result = funcs.reduce_mse_cr(u, measurement)
    # else:
    #     raise ValueError(f"measurement dtype {measurement.dtype} is incompatible")
    result = funcs.reduce_mse(u, measurement)
    if stream is None:
        return result.get()
    else:
        return result.get(pagelocked=True, async_=True, stream=stream)


def reduce_mse_grad(u, measurement, output=None, stream=None):
    param_check(u=u, measurement=measurement, output=output)
    if u.dtype != np.complex128:
        raise ValueError(f"u dtype {u.dtype} is incompatible")
    funcs = get_funcs(u, model="any", stream=stream)
    if output is None:
        output = gpuarray.empty_like(u)
    funcs.mse_grad(u, measurement, output)
    # if measurement.dtype == np.complex128:
    #     funcs.mse_cc_grad(u, measurement, output)
    # elif measurement.dtype == np.double:
    #     funcs.mse_cr_grad(u, measurement, output)
    # else:
    #     raise ValueError(f"measurement dtype {measurement.dtype} is incompatible")
    return output


def sum_batch(u, output=None, stream=None):
    param_check(u=u[0], output=output)
    funcs = get_funcs(u, model="any", stream=stream)
    if output is None:
        funcs.sum_batch(u, u[0])
    else:
        funcs.sum_batch(u, output)


def get_multiplier(shape, res=None, stream=None):
    # funcs = get_funcs(arr_like, model="any", stream=stream)
    if res is None:
        res = global_config.res
    return Multipliers(shape, res, stream)


def u_mul(u, mul, copy=False, stream=None, conj=False):
    funcs = get_funcs(u, model="any", stream=stream)
    if copy:
        out = gpuarray.empty_like(u)
    else:
        out = u

    if isinstance(mul, Number):
        if conj:
            mul = mul.conjugate()
        u._axpbz(mul, 0, out, funcs.stream)
        return out

    # GPUArray * GPUArray
    if len(u.shape) == 3:
        if len(mul.shape) == 3:  # rare case: batch * batch, conj will uses temp memory
            param_check(u=u, mul=mul)
            if conj:
                mul = mul.conj()
            u._elwise_multiply(mul, out, funcs.stream)
            return out
        else:
            param_check(u=u[0], mul=mul)
    else:
        param_check(u=u, mul=mul)
    funcs.op(u, "*", mul, out, name="mul")
    return out


def merge_prop(uf, ub, config=None, copy=False, stream=None):
    param_check(uf=uf, ub=ub)
    funcs = get_funcs(uf, config, model="ssnp", stream=stream)
    af = funcs.fft(uf, copy=copy)
    ab = funcs.fft(ub, copy=copy)
    funcs.merge_prop(af, ab)
    u = funcs.ifft(af)
    u_d = funcs.ifft(ab)
    return u, u_d


def split_prop(u, u_d, config=None, copy=False, stream=None):
    param_check(u=u, u_d=u_d)
    funcs = get_funcs(u, config, model="ssnp", stream=stream)
    a = funcs.fft(u, copy=copy)
    a_d = funcs.fft(u_d, copy=copy)
    funcs.split_prop(a, a_d)
    uf = funcs.ifft(a)
    ud = funcs.ifft(a_d)
    return uf, ud


def merge_grad(ufg, ubg, config=None, copy=False, stream=None):
    param_check(uf_grad=ufg, ub_grad=ubg)
    funcs = get_funcs(ufg, config, model="ssnp", stream=stream)
    afg = funcs.fft(ufg, copy=copy)
    abg = funcs.fft(ubg, copy=copy)
    funcs.merge_grad(afg, abg)
    ug = funcs.ifft(afg)
    u_dg = funcs.ifft(abg)
    return ug, u_dg


def get_funcs(arr_like, config=None, *, model="any", stream=None, fft_type="skcuda"):
    name = model.lower()
    if config is None:
        config = global_config
    if name == "any":
        res = n0 = None
    else:
        res = config.res
        n0 = config.n0
    try:
        model_init = {"ssnp": SSNPFuncs, "bpm": BPMFuncs, "any": Funcs}[name]
    except KeyError:
        raise ValueError(f"unknown model {model}") from None
    return model_init(arr_like, res, n0, stream, fft_type)



from typing import Tuple, Union, List

FFF_P = Union[Tuple[float, float, float], property]
F_P = Union[float, property]


def param_check(**kwargs): ...


def get_stream(ctx): ...


def get_stream_in_current(): ...


class Config:
    res: FFF_P
    xyz: FFF_P
    lambda0: F_P
    n0: F_P
    _callbacks: List[callable, ...]

    def register_updater(self, updater): ...

    def clear_updater(self): ...

    def set(self, **kwargs): ...


config: Config



from pycuda.gpuarray import GPUArray
from pycuda.driver import Stream
from typing import Literal, Optional, List, Union
import numpy as np
from ssnp.utils import Multipliers, Config
from ssnp.funcs import Funcs

G_PRO = Union[GPUArray, property]
ARR = Union[GPUArray, np.ndarray]

class TrackStack:
    def __init__(self, u_num, shape): ...

    def clear(self): ...


class BeamArray:
    dtype = np.complex128
    DERIVATIVE = 0
    BACKWARD = 1
    relation: Literal[0, 1]
    _config: Optional[Config]
    config: Union[Config, property]
    _get_array_times: int
    _u1: GPUArray
    _u2: Optional[GPUArray]
    forward: G_PRO
    backward: Optional[G_PRO]
    field: G_PRO
    derivative: Optional[G_PRO]
    multiplier: Multipliers
    _array_pool: List[GPUArray, ...]
    _tape: list
    ops_number: dict
    _fft_funcs: Funcs
    batch: int
    stream: Stream

    def __init__(self, u1: ARR, u2: ARR = None, relation: Literal[0, 1] = DERIVATIVE, total_ops: int = 0): ...

    def _parse(self, info, dz, n, track: bool): ...

    def ssnp(self, dz, n: GPUArray = None, *, track: bool = False): ...

    def bpm(self, dz, n: GPUArray = None, *, track: bool = False): ...

    def forward_mse_loss(self, measurement: GPUArray, *, track: bool = False): ...

    def n_grad(self, output: GPUArray = None) -> GPUArray: ...

    def binary_pupil(self, na: float): ...

    def mul(self, arr: GPUArray, *, hold: BeamArray = None, track=False): ...

    def a_mul(self, arr: GPUArray, *, hold: BeamArray = None, track=False): ...

    def __imul__(self, other): ...

    @staticmethod
    def _iadd_isub(self: BeamArray, other: BeamArray, add: bool): ...

    def __iadd__(self, other: BeamArray): ...

    def __isub__(self, other: BeamArray): ...

    def split_prop(self): ...

    def merge_prop(self): ...

    def _get_array(self) -> GPUArray: ...

    def recycle_array(self, arr: GPUArray): ...

    @staticmethod
    def _u_setter(u: GPUArray, value, dtype=dtype): ...



"""
SSNP

=====

Features

* Forward model calculation based on CUDA
* Read/write data of common file type
* Gradient calculation
* Image reconstruction with regularization
"""

import sys

if sys.version_info.major < 3:
    raise ImportError("SSNP package only support python3")

import warnings

if sys.version_info < (3, 8):
    warnings.warn("Untested python version. Please use python>=3.8")

VERSION = '0.0.2b'

import ssnp.utils.init_helper

from ssnp.beam import BeamArray
from ssnp.utils import Multipliers, config, read, write



from pycuda import elementwise, gpuarray, reduction
from pycuda.gpuarray import GPUArray
from pycuda.tools import dtype_to_ctype
import numpy as np
from ssnp.utils import Multipliers, get_stream_in_current
from contextlib import contextmanager
from functools import partial, lru_cache
from skcuda import fft as skfft


class Funcs:
    # __temp_memory_pool = {}
    _funcs_cache = {}
    reduce_mse_cr_krn = None

    def __new__(cls, arr_like, res, n0, stream=None, fft_type="reikna"):
        if cls.reduce_mse_cr_krn is None:
            Funcs.reduce_mse_cr_krn = reduction.ReductionKernel(
                dtype_out=np.double, neutral=0,
                reduce_expr="a+b",
                map_expr="(cuCabs(x[i]) - y[i]) * (cuCabs(x[i]) - y[i])",
                arguments="double2 *x, double *y",
                preamble='#include "cuComplex.h"'
            )
            Funcs.reduce_mse_cc_krn = reduction.ReductionKernel(
                dtype_out=np.double, neutral=0,
                reduce_expr="a+b",
                map_expr="cuCabs(cuCsub(x[i], y[i])) * cuCabs(cuCsub(x[i], y[i]))",
                arguments="double2 *x, double2 *y",
                preamble='#include "cuComplex.h"'
            )
            Funcs.mse_cc_grad_krn = elementwise.ElementwiseKernel(
                "double2 *u, double2 *m, double2 *out",
                """
                out[i] = cuCsub(u[i], m[i]);
                out[i].x *= 2; out[i].y *= 2;
                """,
                preamble='#include "cuComplex.h"'
            )
            Funcs.mse_cr_grad_krn = elementwise.ElementwiseKernel(
                "double2 *u, double *m, double2 *out",
                """
                temp = 2 * (1 - m[i] / cuCabs(u[i]));
                out[i].x = temp * u[i].x; out[i].y = temp * u[i].y;
                """,
                loop_prep="double temp",
                preamble='#include "cuComplex.h"'
            )

        shape = tuple(arr_like.shape)
        key = (shape, res, stream, fft_type)
        try:
            return cls._funcs_cache[key]
        except KeyError:
            cls._funcs_cache[key] = super().__new__(cls)
            return cls._funcs_cache[key]

    def __init__(self, arr_like, res, n0, stream=None, fft_type="reikna"):
        if self._funcs_cache is None:
            return
        self._funcs_cache = None  # only for mark as initialized

        if stream is None:
            stream = get_stream_in_current()
        self.stream = stream

        # FFTs
        shape = tuple(arr_like.shape)
        if fft_type == "reikna":
            if len(shape) != 2:
                raise NotImplementedError(f"cannot process {len(shape)}-D data with reikna")
            self._fft_callable = self._compile_fft(arr_like.shape, arr_like.dtype, stream)
            self.fft = self._fft_reikna
            batch = 1
        elif fft_type == "skcuda":
            if len(shape) == 3:
                batch = shape[0]
                shape = shape[1:]
            else:
                batch = 1
            if len(shape) != 2:
                raise NotImplementedError(f"cannot process {len(shape)}-D data with skcuda")
            self._skfft_plan = skfft.Plan(shape, np.complex128, np.complex128, batch, stream)
            self.fft = self._fft_sk
        else:
            raise ValueError(f"Unknown FFT type {fft_type}")
        self.ifft = partial(self.fft, inverse=True)
        self.shape = shape
        self.batch = batch

        # kz and related multipliers
        if type(self) is not Funcs:
            self._prop_cache = {}
            self._pupil_cache = {}
            self.res = res
            self.n0 = n0
            self.multiplier = Multipliers(shape, res, stream)
            c_gamma = self.multiplier.c_gamma()
            c_gamma = np.broadcast_to(c_gamma, arr_like.shape)
            kz = c_gamma * (2 * np.pi * res[2])
            self.kz = kz.astype(np.double)
            self.kz_gpu = gpuarray.to_gpu(kz)
            self.eva = np.exp(np.minimum((c_gamma - 0.2) * 5, 0))

        # function interface
        self.sum_cmplx_batch_krn = elementwise.ElementwiseKernel(
            "double2 *out, double2 *u",
            f"""
            out[i] = u[i];
            for (j = 1; j < {self.batch}; j++) {{
                out[i].x += u[i + j * n].x;
                out[i].y += u[i + j * n].y;
            }}
            """,
            loop_prep="unsigned j",
        )
        self.sum_double_batch_krn = elementwise.ElementwiseKernel(
            "double *out, double *u",
            f"""
            out[i] = u[i];
            for (j = 1; j < {self.batch}; j++)
                out[i] += u[i + j * n];
            """,
            loop_prep="unsigned j",
        )
        # self.reduce_mse_cr = partial(self.reduce_mse_cr_krn, stream=stream)
        # self.reduce_mse_cc = partial(self.reduce_mse_cc_krn, stream=stream)
        # self.mse_cc_grad = partial(self.mse_cc_grad_krn, stream=stream)
        # self.mse_cr_grad = partial(self.mse_cr_grad_krn, stream=stream)

    @staticmethod
    @lru_cache
    def op_krn(batch, xt, yt, zt, operator, name=None, y_func=None):
        y_i = f"y[i % (n / {batch})]"
        if y_func is not None:
            y_i = f"{y_func}(y_i)"
        return elementwise.get_elwise_kernel(
            f"{xt} *x, {yt} *y, {zt} *z",
            f"z[i] = x[i] {operator} {y_i}",
            name or "op")

    def op(self, x, operator, y, out=None, **kwargs):
        if out is None:
            out = x
        xt = dtype_to_ctype(x.dtype)
        yt = dtype_to_ctype(y.dtype)
        zt = dtype_to_ctype(out.dtype)
        func = self.op_krn(self.batch, xt, yt, zt, operator, **kwargs)
        func.prepared_async_call(x._grid, x._block, self.stream,
                                 x.gpudata, y.gpudata,
                                 out.gpudata, x.mem_size)
        return out

    def reduce_mse(self, field, measurement):
        if field.dtype == np.complex128:
            if measurement.dtype == np.float64:
                return Funcs.reduce_mse_cr_krn(field, measurement, stream=self.stream)
            elif measurement.dtype == np.complex128:
                return Funcs.reduce_mse_cc_krn(field, measurement, stream=self.stream)
        raise TypeError(f"incompatible dtype: field {field.dtype}, measurement {measurement.dtype}")

    def mse_grad(self, field, measurement, gradient):
        if field.dtype == np.complex128 and gradient.dtype == np.complex128:
            if measurement.dtype == np.float64:
                return Funcs.mse_cr_grad_krn(field, measurement, gradient, stream=self.stream)
            elif measurement.dtype == np.complex128:
                return Funcs.mse_cc_grad_krn(field, measurement, gradient, stream=self.stream)
        raise TypeError(f"incompatible dtype: field {field.dtype}, measurement {measurement.dtype}, "
                        f"gradient {gradient.dtype}")

    def sum_batch(self, batch, sum_):
        if batch.dtype == np.complex128 and sum_.dtype == np.complex128:
            return self.sum_cmplx_batch_krn(sum_, batch, stream=self.stream)
        elif batch.dtype == np.float64 and sum_.dtype == np.float64:
            return self.sum_double_batch_krn(sum_, batch, stream=self.stream)
        raise TypeError(f"incompatible dtype: field {batch.dtype}, measurement {sum_.dtype}")

    @staticmethod
    @lru_cache
    def _compile_fft(shape, dtype, stream):
        from reikna.fft import FFT
        import reikna.cluda as cluda
        thr = cluda.cuda_api().Thread(stream)
        arr_like = type("", (), {})()
        arr_like.shape = shape
        arr_like.dtype = dtype
        return FFT(arr_like).compile(thr)

    def _fft_reikna(self, arr, output=None, copy=False, inverse=False):
        if output is not None:
            o = output
        elif copy:
            o = gpuarray.empty_like(arr)
        else:
            o = arr
        self._fft_callable(o, arr, inverse=inverse)
        return o

    def _fft_sk(self, arr, *, output=None, copy=False, inverse=False):
        if output is not None:
            o = output
        elif copy:
            o = gpuarray.empty_like(arr)
        else:
            o = arr
        fft = skfft.ifft if inverse else skfft.fft
        fft(arr, o, self._skfft_plan)
        if inverse:
            scale = 1 / self.shape[-1] / self.shape[-2]
            o._axpbz(scale, 0, o, stream=self.stream)
        return o

    @contextmanager
    def fourier(self, arr, copy=False):
        yield self.fft(arr, copy=copy)
        self.ifft(arr)

    def diffract(self, *args):
        raise NotImplementedError

    def scatter(self, *args):
        raise NotImplementedError

    def diffract_g(self, ag, dz):
        raise NotImplementedError

    def scatter_g(self, u, n, ug, ng, dz):
        raise NotImplementedError

    def _get_prop(self, dz):
        """
            P: 2D angular spectrum transmittance numpy array(s). Evanescent wave removed:
            For 0.98<N.A.<1, decrease to 1~1/e per step;
            For N.A.>1 decrease to 1/e per step

            Q: scattering operation, different between SSNP & BPM

            Pg/Qg: transposed array/operation of P/Q

            :return: Dictionary of P, Q, Pg, Qg
        """
        raise NotImplementedError

    def conj(self, arr: GPUArray):
        """copy from GPUArray.conj(self), do conj in-place"""
        dtype = arr.dtype
        if not arr.flags.forc:
            raise RuntimeError("only contiguous arrays may "
                               "be used as arguments to this operation")
        func = elementwise.get_conj_kernel(dtype)
        func.prepared_async_call(arr._grid, arr._block, self.stream,
                                 arr.gpudata, arr.gpudata, arr.mem_size)

    # @staticmethod
    # def get_temp_mem(arr_like: GPUArray, index=0):
    #     key = (arr_like.shape, arr_like.dtype, index)
    #     try:
    #         return Funcs.__temp_memory_pool[key]
    #     except KeyError:
    #         mem = gpuarray.empty_like(arr_like)
    #         Funcs.__temp_memory_pool[key] = mem
    #         return mem


class SSNPFuncs(Funcs):
    _funcs_cache = {}
    _fused_mam_callable_krn = elementwise.ElementwiseKernel(
        "double2 *a, double2 *b, "
        "double2 *p00, double2 *p01, double2 *p10, double2 *p11",
        """
        temp = cuCfma(a[i], p00[i], cuCmul(b[i], p01[i]));
        b[i] = cuCfma(a[i], p10[i], cuCmul(b[i], p11[i]));
        a[i] = temp;
        """,
        loop_prep="cuDoubleComplex temp",
        preamble='#include "cuComplex.h"',
        name='fused_mam'
    )
    _split_prop_krn = elementwise.ElementwiseKernel(
        # ab = (a + I a_d / kz) / 2
        # af = a - ab = (a - I a_d / kz) / 2
        "double2 *a, double2 *a_d, double *kz",
        """
        temp.x = (cuCreal(a[i]) - cuCimag(a_d[i])/kz[i])*0.5;
        temp.y = (cuCimag(a[i]) + cuCreal(a_d[i])/kz[i])*0.5;
        a_d[i] = temp;
        a[i] = cuCsub(a[i], a_d[i]);
        """,
        loop_prep="cuDoubleComplex temp",
        preamble='#include "cuComplex.h"'
    )
    _merge_prop_krn = elementwise.ElementwiseKernel(
        # a = af + ab
        # a_d = (af - ab) * I kz
        "double2 *af, double2 *ab, double *kz",
        """
        temp = cuCsub(af[i], ab[i]);
        af[i] = cuCadd(af[i], ab[i]);
        ab[i] = make_cuDoubleComplex(-cuCimag(temp)*kz[i], cuCreal(temp)*kz[i]);
        """,
        loop_prep="cuDoubleComplex temp",
        preamble='#include "cuComplex.h"'
    )
    _merge_grad_krn = elementwise.ElementwiseKernel(
        # ag = (afg + abg) / 2
        # a_dg = (afg - abg) / 2 * I / kz
        "double2 *afg, double2 *abg, double *kz",
        """
        afg[i].x *= 0.5; afg[i].y *= 0.5;
        abg[i].x *= 0.5; abg[i].y *= 0.5;
        temp = cuCsub(afg[i], abg[i]);
        afg[i] = cuCadd(afg[i], abg[i]);
        abg[i] = make_cuDoubleComplex(-cuCimag(temp)/kz[i], cuCreal(temp)/kz[i]);
        """,
        loop_prep="cuDoubleComplex temp",
        preamble='#include "cuComplex.h"'
    )

    def _get_prop(self, dz):
        res = self.res
        n0 = self.n0
        key = round(dz * 1000)
        try:
            return self._prop_cache[key]
        except KeyError:
            kz = self.kz
            eva = self.eva
            p_mat = [np.cos(kz * dz), np.sin(kz * dz) / kz,
                     -np.sin(kz * dz) * kz, np.cos(kz * dz)]
            p_mat = [gpuarray.to_gpu((i * eva).astype(np.complex128)) for i in p_mat]
            phase_factor = (2 * np.pi * res[2]) ** 2 * dz
            q_op = elementwise.ElementwiseKernel(
                "double2 *u, double2 *ud, double *n_",
                f"""
                    temp = n_[i % (n / {self.batch})];
                    temp = {phase_factor / n0 ** 2} * temp * ({2 * n0} + temp);
                    ud[i].x -= temp * u[i].x;
                    ud[i].y -= temp * u[i].y;
                """,
                loop_prep="double temp",
                name="ssnp_q"
            )
            q_op_g = elementwise.ElementwiseKernel(
                "double2 *u, double *n_, double2 *ug, double2 *udg, double *n_g",
                # Forward: ud = ud - temp * u
                # ng = Re{-udg * conj(u) * (d_temp(n) / d_n)}
                f"""
                    ni = n_[i % (n / {self.batch})];
                    temp = {phase_factor} * ni * ({2 * n0} + ni);
                    ug[i].x -= temp * udg[i].x;
                    ug[i].y -= temp * udg[i].y;
                    n_g[i] = -(udg[i].x * u[i].x + udg[i].y * u[i].y) * {phase_factor} * ({2 * n0} + 2 * ni)
                """,
                loop_prep="double temp, ni",
                preamble='#include "cuComplex.h"',
                name="ssnp_qg"
            )
            new_prop = {"P": p_mat, "Pg": [p_mat[i].conj() for i in (0, 2, 1, 3)], "Q": q_op, "Qg": q_op_g}
            self._prop_cache[key] = new_prop
            return new_prop

    def diffract(self, a, a_d, dz):
        self._fused_mam_callable_krn(a, a_d, *self._get_prop(dz)["P"], stream=self.stream)

    def scatter(self, u, u_d, n, dz):
        self._get_prop(dz)["Q"](u, u_d, n, stream=self.stream)

    def diffract_g(self, ag, a_dg, dz):
        self._fused_mam_callable_krn(ag, a_dg, *self._get_prop(dz)["Pg"], stream=self.stream)

    def scatter_g(self, u, n, ug, u_dg, ng, dz):
        self._get_prop(dz)["Qg"](u, n, ug, u_dg, ng, stream=self.stream)

    def merge_prop(self, af, ab):
        self._merge_prop_krn(af, ab, self.kz_gpu, stream=self.stream)

    def split_prop(self, a, a_d):
        self._split_prop_krn(a, a_d, self.kz_gpu, stream=self.stream)

    def merge_grad(self, afg, abg):
        self._merge_grad_krn(afg, abg, self.kz_gpu, stream=self.stream)

    def split_grad(self, ag, a_dg):
        raise NotImplementedError


class BPMFuncs(Funcs):
    _funcs_cache = {}

    def _get_prop(self, dz):
        res = self.res
        n0 = self.n0
        key = (round(dz * 1000), "bpm")
        try:
            return self._prop_cache[key]
        except KeyError:
            kz = self.kz.astype(np.complex128)
            eva = self.eva
            p_mat = np.exp(kz * (1j * dz))
            p_mat = gpuarray.to_gpu(p_mat * eva)
            phase_factor = 2 * np.pi * res[2] * n0 * dz
            q_op = elementwise.ElementwiseKernel(
                "double2 *u, double *n_",
                f"""
                    ni = n_[i % (n / {self.batch})];
                    temp = make_cuDoubleComplex(cos(ni * {phase_factor}), sin(ni * {phase_factor}));
                    u[i] = cuCmul(u[i], temp);
                """,
                loop_prep="double2 temp, double ni",
                preamble='#include "cuComplex.h"'
            )
            q_op_g = elementwise.ElementwiseKernel(
                "double2 *u, double *n_, double2 *ug, double *n_g",
                # Forward: u=u*temp
                f"""
                    temp = n_[i % (n / {self.batch})];
                    temp_conj = make_cuDoubleComplex(cos(temp * {phase_factor}), -sin(temp * {phase_factor}));
                    n_g[i] = {phase_factor} * 
                        (cuCimag(ug[i]) * cuCreal(u[i]) - cuCimag(u[i]) * cuCreal(ug[i]));
                    ug[i] = cuCmul(ug[i], temp_conj);
                """,
                loop_prep="double2 temp_conj",
                preamble='#include "cuComplex.h"'
            )
            new_prop = {"P": p_mat, "Pg": p_mat.conj(), "Q": q_op, "Qg": q_op_g}
            self._prop_cache[key] = new_prop
            return new_prop

    def diffract(self, a, dz):
        a._elwise_multiply(self._get_prop(dz)["P"], a, stream=self.stream)
        # a *= self._get_prop(dz)["P"]

    def diffract_g(self, ag, dz):
        ag._elwise_multiply(self._get_prop(dz)["Pg"], ag, stream=self.stream)
        # ag *= self._get_prop(dz)["Pg"]

    def scatter(self, u, n, dz):
        self._get_prop(dz)["Q"](u, n, stream=self.stream)

    def scatter_g(self, u, n, ug, ng, dz):
        self._get_prop(dz)["Qg"](u, n, ug, ng, stream=self.stream)



from typing import Literal

import numpy as np
from pycuda.gpuarray import GPUArray
from pycuda.elementwise import ElementwiseKernel
from pycuda.driver import Stream, Function
from pycuda.reduction import ReductionKernel
from reikna.core.computation import ComputationCallable
from .utils import Multipliers


class Funcs:
    shape: tuple
    batch: int
    res: tuple
    n0: float
    stream: Stream
    kz: np.ndarray
    kz_gpu: GPUArray
    eva: np.ndarray
    multiplier: Multipliers
    _fft_callable: ComputationCallable
    # __temp_memory_pool: dict
    _prop_cache: dict
    reduce_mse_cr_krn: ReductionKernel
    reduce_mse_cc_krn: ReductionKernel
    mse_cr_grad_krn: ElementwiseKernel
    mse_cc_grad_krn: ElementwiseKernel
    mul_conj_krn: ElementwiseKernel
    mul_krn: ElementwiseKernel
    sum_cmplx_batch_krn: ElementwiseKernel
    sum_double_batch_krn: ElementwiseKernel
    _fft_reikna: callable
    _fft_sk: callable

    def __init__(self, arr_like: GPUArray, res, n0, stream: Stream = None,
                 fft_type: Literal["reikna", "skcuda"] = "reikna"): ...

    @staticmethod
    def _compile_fft(shape, dtype, stream): ...

    def fft(self, arr: GPUArray, output: GPUArray = None, copy: bool = False, inverse=False) -> GPUArray: ...

    def ifft(self, arr: GPUArray, output: GPUArray = None, copy: bool = False) -> GPUArray: ...

    def fourier(self, arr: GPUArray, copy: bool = False): ...

    def diffract(self, *args) -> None: ...

    def scatter(self, *args) -> None: ...

    def diffract_g(self, *args) -> None: ...

    def scatter_g(self, *args) -> None: ...

    def _get_prop(self, dz): ...

    # @staticmethod
    # def get_temp_mem(arr_like: GPUArray, index=0): ...

    # def reduce_mse_cr(self, u: GPUArray, m: GPUArray) -> GPUArray: ...
    # def reduce_mse_cc(self, u: GPUArray, m: GPUArray) -> GPUArray: ...
    # def mse_cr_grad(self, u: GPUArray, m: GPUArray, out: GPUArray): ...
    # def mse_cc_grad(self, u: GPUArray, m: GPUArray, out: GPUArray): ...

    def reduce_mse(self, field: GPUArray, measurement: GPUArray) -> GPUArray: ...

    def mse_grad(self, field: GPUArray, measurement: GPUArray, gradient: GPUArray): ...

    def sum_batch(self, batch: GPUArray, sum_: GPUArray): ...

    @staticmethod
    def op_krn(*args, **kwargs) -> Function: ...

    def op(self, x: GPUArray, operator: Literal["+", "-", "*", "/"], y: GPUArray, out: GPUArray = None,
           name: str = None, y_func: str = None): ...


class BPMFuncs(Funcs):
    def _get_prop(self, dz): ...

    def diffract(self, a: GPUArray, dz) -> None: ...

    def diffract_g(self, ag, dz): ...

    def scatter(self, u, n, dz) -> None: ...

    def scatter_g(self, u, n, ug, ng, dz): ...


class SSNPFuncs(Funcs):
    _fused_mam_callable_krn: ElementwiseKernel
    _merge_prop_krn: ElementwiseKernel
    _split_prop_krn: ElementwiseKernel
    _merge_grad_krn: ElementwiseKernel

    def merge_prop(self, af, ab): ...

    def split_prop(self, a, a_d): ...

    def merge_grad(self, afg, abg): ...

    def split_grad(self, ag, a_dg): ...

    def _get_prop(self, dz): ...

    def diffract(self, a, a_d, dz) -> None: ...

    def diffract_g(self, ag, a_dg, dz): ...

    def scatter(self, u, u_d, n, dz) -> None: ...

    def scatter_g(self, u, n, ug, u_dg, ng, dz): ...



from typing import Callable, overload, Literal, Tuple, Optional
from numbers import Real
from numpy import ndarray
from pycuda.gpuarray import GPUArray
from pycuda.driver import Stream


def _cache_array(func: Callable[..., Tuple[tuple, Callable[[], ndarray]]]): ...


class Multipliers:
    _cache: dict
    _gpu_cache: dict
    _xy_size: Tuple[int, int]
    _shape: Tuple[int, int]
    res: Tuple[float, float, float]
    stream: Optional[Stream]

    def __init__(self, shape: Tuple[int, int], res: Tuple[float, float, float], stream: Stream = None): ...

    @overload
    def tilt(self, c_ab, *, trunc, periodic_params=None, c_ab_out=None, gpu: Literal[True]) -> GPUArray: ...

    @overload
    def tilt(self, c_ab, *, trunc, periodic_params=None, c_ab_out=None, gpu: Literal[False] = False) -> ndarray: ...

    @overload
    def binary_pupil(self, na, *, gpu: Literal[True]) -> GPUArray: ...

    @overload
    def binary_pupil(self, na, *, gpu: Literal[False] = False) -> ndarray: ...

    @overload
    def c_gamma(self, *, gpu: Literal[True]) -> GPUArray: ...

    @overload
    def c_gamma(self, *, gpu: Literal[False] = False) -> ndarray: ...

    @staticmethod
    def _near_0(size, pos_0): ...

    @overload
    def soft_crop(self, width: Real, *, total_slices: int = 1, pos: Real = 0, strength: Real = 1,
                  gpu: Literal[True]) -> GPUArray: ...

    @overload
    def soft_crop(self, width: Real, *, total_slices: int = 1, pos: Real = 0, strength: Real = 1,
                  gpu: Literal[False] = False) -> ndarray: ...

    @overload
    def hard_crop(self, width, gpu: Literal[True]) -> GPUArray: ...

    @overload
    def hard_crop(self, width, gpu: Literal[False] = False) -> ndarray: ...

    @overload
    def gaussian(self, sigma: Real, mu: Tuple[Real, Real] = (0, 0), *, gpu: Literal[True]) -> GPUArray: ...

    @overload
    def gaussian(self, sigma: Real, mu: Tuple[Real, Real] = (0, 0), *, gpu: Literal[False] = False) -> ndarray: ...
